from enum import StrEnum


class ModelEnum(StrEnum):
    UNKNOWN = "Unknown"
    HUMIDIFIER_H35I = "Blueair Humidifier H35i"
    PROTECT_7440I = "Blueair Protect 7440i"
    PROTECT_7470I = "Blueair Protect 7470i"
    MAX_211I = "Blueair Blue Pure 211i Max"
    MAX_311I = "Blueair Blue Pure 311i Max"
    MAX_311I_PLUS = "Blueair Blue Pure 311i+ Max"
    MAX_3250I = "Blueair Blue Pure 3250i Max"
    MAX_3650I = "Blueair Blue Pure 3650i Max"
    MAX_411I = "Blueair Blue Pure 411i Max"
    MAX_511I = "Blueair Blue Pure 511i Max"
    T10I = "T10i ComfortPure 3-in-1 Filter/Heater/Fan"
    BLUE_SIGNATURE = "Blue Signature"
    PET_AIR_PRO = "PetAir Pro"
